﻿using System;
using System.Diagnostics;
using System.IO;
using System.Text;
using System.Text.RegularExpressions;
using Microsoft.Win32;

namespace Minemapper
{
	public class JavaChecker
	{
		private static Regex _javaVersionRegex = new Regex("java version \"(\\d*\\.\\d*)\\.", RegexOptions.CultureInvariant | RegexOptions.Compiled);
		private static bool _haveCheckedRegistry = false;
		private static bool _haveCheckedRegistryWow6432Node = false;

		public static string JavaFileName { get; set; }

		private JavaChecker()
		{
		}

		public static bool EnsureJava(Version minimumRequired)
		{
			try
			{
				Version version = GetJavaVersion();

				return version >= minimumRequired;
			}
			catch (Exception ex)
			{
				Debug.WriteLine(ex);
			}

			return false;
		}

		private static string GetJavaFileNameFromRegistry()
		{
			string currentVersion = Registry.GetValue(@"HKEY_LOCAL_MACHINE\Software\JavaSoft\Java Runtime Environment", "CurrentVersion", "") as string;
			if (string.IsNullOrEmpty(currentVersion))
				return string.Empty;

			string javaHome = Registry.GetValue(@"HKEY_LOCAL_MACHINE\Software\JavaSoft\Java Runtime Environment\" + currentVersion, "JavaHome", "") as string;
			if (string.IsNullOrEmpty(javaHome))
				return string.Empty;

			string javaFileName = Path.Combine(javaHome, @"bin\java.exe");

			return javaFileName;
		}

		private static string GetJavaFileNameFromRegistryWow6432Node()
		{
			string currentVersion = Registry.GetValue(@"HKEY_LOCAL_MACHINE\Software\Wow6432Node\JavaSoft\Java Runtime Environment", "CurrentVersion", "") as string;
			if (string.IsNullOrEmpty(currentVersion))
				return string.Empty;

			string javaHome = Registry.GetValue(@"HKEY_LOCAL_MACHINE\Software\Wow6432Node\JavaSoft\Java Runtime Environment\" + currentVersion, "JavaHome", "") as string;
			if (string.IsNullOrEmpty(javaHome))
				return string.Empty;

			string javaFileName = Path.Combine(javaHome, @"bin\java.exe");

			return javaFileName;
		}

		public static Version GetJavaVersion()
		{
			_haveCheckedRegistry = false;
			_haveCheckedRegistryWow6432Node = false;

			string output = GetJavaVersionText();

			Match m = _javaVersionRegex.Match(output);

			string errorMessage = string.Format("Execution of '{0} -version' failed to return a valid version", JavaFileName);
			if (!m.Success)
				throw new JavaNotDetectedException(errorMessage);

			if (m.Groups == null || m.Groups.Count < 2)
				throw new JavaNotDetectedException(errorMessage);

			output = m.Groups[1].Value;

			Version version = new Version(output);

			return version;
		}

		public static string GetJavaVersionText()
		{
			try
			{
				ProcessStartInfo info = new ProcessStartInfo(JavaFileName, "-version");
				info.UseShellExecute = false;
				info.CreateNoWindow = true;
				info.WindowStyle = ProcessWindowStyle.Hidden;
				info.RedirectStandardOutput = true;
				info.RedirectStandardError = true;

				Process process = new Process();
				process.StartInfo = info;
				process.EnableRaisingEvents = true;

				StringBuilder output = new StringBuilder();

				process.OutputDataReceived += delegate(object sender, DataReceivedEventArgs e)
				{
					Debug.WriteLine(e.Data);
					output.AppendLine(e.Data);
				};

				process.ErrorDataReceived += delegate(object sender, DataReceivedEventArgs e)
				{
					Debug.WriteLine(e.Data);
					output.AppendLine(e.Data);
				};

				process.Start();
				process.BeginOutputReadLine();
				process.BeginErrorReadLine();
				process.WaitForExit();

				return output.ToString();
			}
			catch (Exception ex)
			{
				Debug.WriteLine(ex);

				if (!_haveCheckedRegistry)
				{
					// couldn't find java.exe at the path specified by Properties.Settings.Default.JavaFileName,
					// or it's not in the PATH environment variable.  Check the registry.
					_haveCheckedRegistry = true;

					string javaFileName = GetJavaFileNameFromRegistry();

					JavaFileName = javaFileName;

					return GetJavaVersionText();
				}
				else if (!_haveCheckedRegistryWow6432Node)
				{
					_haveCheckedRegistryWow6432Node = true;

					string javaFileName = GetJavaFileNameFromRegistryWow6432Node();
					if (string.IsNullOrEmpty(javaFileName) || !File.Exists(javaFileName))
					{
						Debug.WriteLine("Could not find java filename in the Windows Registry Wow6432Node.");
						return string.Empty;
					}

					JavaFileName = javaFileName;

					return GetJavaVersionText();
				}
			}

			return string.Empty;
		}
	}

	public class JavaNotDetectedException : Exception
	{
		public JavaNotDetectedException() { }
		public JavaNotDetectedException(string message) : base(message) { }
		public JavaNotDetectedException(string message, Exception inner) : base(message, inner) { }
	}
}
