﻿using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.IO;
using System.Linq;
using System.Threading.Tasks;
using System.Windows;
using System.Windows.Controls;
using System.Windows.Input;
using System.Windows.Media;
using System.Windows.Media.Animation;
using System.Windows.Media.Imaging;
using System.Windows.Threading;
using log4net;
using Microsoft.Sample.Controls;
using Microsoft.Win32;

namespace Minemapper
{
	public partial class MainWindow : Window
	{
		#region Variables

		private static readonly ILog logger = LogManager.GetLogger(typeof(MainWindow));

		string applicationData;

		int lastTick = Environment.TickCount;
		int addedPerSecond = 0;
		int removedPerSecond = 0;

		private bool ignoreDirectionCheck = true;
		private string direction = "east";
		private bool biomes = true;
		private bool night = false;
		private bool skylight = false;
		private bool blendAll = false;
		private bool nether = false;
		private bool caves = false;

		private int worldHeight;

		bool ignoreWorldCheck = false;
		bool ignoreSliderValueChanges = false;
		bool ignoreScaleChanges = false;

		private DirectoryInfo currentWorldDirectoryInfo;
		private DirectoryInfo currentImageDirectoryInfo;

		//private SelectionImage selection;

		TileImageCreator tileImageCreator = new TileImageCreator();
		private bool ignoreModeCheck;

		#endregion

		#region Constructors

		public MainWindow()
		{
			logger.Info("MainWindow constructor");

			InitializeComponent();

			this.Loaded += new RoutedEventHandler(MainWindow_Loaded);

			this.virtualCanvas.ScaleTransform.Changed += new EventHandler(ScaleTransform_Changed);

			tileImageCreator.TilesCreated += this.tileImageCreator_TilesCreated;
			tileImageCreator.TileImageLoaded += this.tileImageCreator_TileImageLoaded;
			tileImageCreator.StatusChanged += this.tileImageCreator_StatusChanged;
			tileImageCreator.ProgressChanged += this.tileImageCreator_ProgressChanged;
			tileImageCreator.ProcessCountChanged += this.tileImageCreator_ProcessCountChanged;
		}

		#endregion

		#region Methods

		private void LoadMinecraftWorldMenuItems()
		{
			this.menuItemWorlds.Items.Clear();

			string savesPath = Path.Combine(applicationData, ".minecraft", "saves");

			DirectoryInfo directoryInfo = new DirectoryInfo(savesPath);

			if (directoryInfo.Exists)
			{
				foreach (DirectoryInfo worldDirectoryInfo in directoryInfo.GetDirectories())
				{
					MenuItem menuItem = new MenuItem();

					menuItem.Header = worldDirectoryInfo.Name;
					menuItem.Tag = worldDirectoryInfo;
					menuItem.IsCheckable = true;
					menuItem.Checked += new RoutedEventHandler(OnLoadWorldMenuItem_Checked);

					this.menuItemWorlds.Items.Add(menuItem);
				}

				this.menuItemWorlds.Items.Add(new Separator());
			}

			MenuItem menuItemBrowseForWorld = new MenuItem();
			menuItemBrowseForWorld.Header = "Browse...";
			menuItemBrowseForWorld.Click += new RoutedEventHandler(menuItemBrowseForWorld_Click);
			this.menuItemWorlds.Items.Add(menuItemBrowseForWorld);

			this.menuItemWorlds.Items.Add(new Separator());

			MenuItem menuItemRefresh = new MenuItem();
			menuItemRefresh.Header = "Refresh";
			menuItemRefresh.Click += new RoutedEventHandler(menuItemRefresh_Click);
			this.menuItemWorlds.Items.Add(menuItemRefresh);

			MenuItem menuItemViewCacheFolder = new MenuItem();
			menuItemViewCacheFolder.Header = "View Cache Folder";
			menuItemViewCacheFolder.Click += new RoutedEventHandler(menuItemViewCacheFolder_Click);
			this.menuItemWorlds.Items.Add(menuItemViewCacheFolder);
		}

		private void ReloadWorld()
		{
			if (this.currentWorldDirectoryInfo == null)
				return;

			this.LoadWorld(this.currentWorldDirectoryInfo);
		}

		private void LoadWorld(DirectoryInfo worldDirectoryInfo)
		{
			StatusText.Text = "Clearing current display...";

			this.Clear();

			this.Title = "Minemapper - " + worldDirectoryInfo.Name;

			this.currentWorldDirectoryInfo = worldDirectoryInfo;
			string imagePath = Path.Combine(worldDirectoryInfo.Name + "-ImageCache", this.direction);

			if (this.night)
				imagePath = Path.Combine(imagePath, "Night");
			else if (this.caves)
				imagePath = Path.Combine(imagePath, "Caves");
			else if (this.nether)
				imagePath = Path.Combine(imagePath, "Nether");
			else
				imagePath = Path.Combine(imagePath, "Day");

			if (this.skylight)
				imagePath = Path.Combine(imagePath, "Skylight");
			else
				imagePath = Path.Combine(imagePath, "NoSkylight");

			if (this.biomes)
				imagePath = Path.Combine(imagePath, "Biomes");
			else
				imagePath = Path.Combine(imagePath, "NoBiomes");

			imagePath = Path.Combine(Environment.GetFolderPath(Environment.SpecialFolder.ApplicationData), "Minemapper", imagePath);

			bool canUseCache = currentImageDirectoryInfo != null && string.Compare(imagePath, currentImageDirectoryInfo.FullName, true) != 0;

			DirectoryInfo imageDirectoryInfo = new DirectoryInfo(imagePath);
			this.currentImageDirectoryInfo = imageDirectoryInfo;

			if (canUseCache && this.currentImageDirectoryInfo.Exists)
			{
				string message = string.Format(
					"Minemapper has found previously cached images for '{0}'.  Minemapper can either reuse these cached images or clear them.  Reusing the cached images will make loading much faster, but will not show changes to the world since the cached images were first generated.\n\nWould you like to use the cached images?",
					this.currentWorldDirectoryInfo.Name);

				MessageBoxResult result = MessageBox.Show(this, message, "Reuse cached images?", MessageBoxButton.YesNo, MessageBoxImage.Question, MessageBoxResult.Yes);
				if (result == MessageBoxResult.No)
				{
					try
					{
						Directory.Delete(this.currentImageDirectoryInfo.FullName, true);
					}
					catch (Exception ex)
					{
						this.HandleException(ex);
					}
				}
			}

			this.SetWorldControlsState(false);
			this.menuItemSave.IsEnabled = true;
			this.StatusBar.Visibility = Visibility.Visible;

			this.ignoreWorldCheck = true;

			try
			{
				foreach (MenuItem menuItem in this.menuItemWorlds.Items.OfType<MenuItem>())
				{
					DirectoryInfo other = menuItem.Tag as DirectoryInfo;
					if (other == null)
						continue;

					if (string.Compare(other.FullName, worldDirectoryInfo.FullName, true) == 0)
					{
						if (!menuItem.IsChecked)
							menuItem.IsChecked = true;
					}
					else if (menuItem.IsChecked)
					{
						menuItem.IsChecked = false;
					}
				}
			}
			finally
			{
				this.ignoreWorldCheck = false;
			}

			Action<DirectoryInfo> action = new Action<DirectoryInfo>(this.LoadTilesAsync);

			Task.Factory.StartNew(() => action(worldDirectoryInfo));
		}

		private void SetWorldControlsState(bool state)
		{
			this.menuItemWorlds.IsEnabled = state;
			this.menuItemDirection.IsEnabled = state;
			this.comboBoxMode.IsEnabled = state;
			this.menuItemBiomes.IsEnabled = state;
			this.menuItemBlendAll.IsEnabled = state;
			this.menuItemSkylight.IsEnabled = state;
			this.stackPanelHeight.IsEnabled = state;
		}

		private void Clear()
		{
			this.virtualCanvas.VirtualChildren.Clear();

			//this.selection = new SelectionImage();

			//RenderOptions.SetBitmapScalingMode(this.selection, BitmapScalingMode.HighQuality);
			//RenderOptions.SetEdgeMode(this.selection, EdgeMode.Unspecified);

			//string path = Path.GetDirectoryName(System.Reflection.Assembly.GetExecutingAssembly().Location);
			//path = Path.Combine(path, "selection.png");
			//Uri selectionImageUri = new Uri(path);

			//this.selection.Source = new BitmapImage(selectionImageUri);

			//this.virtualCanvas.ContentCanvas.Children.Add(this.selection);
		}

		void LoadTilesAsync(DirectoryInfo worldDirectoryInfo)
		{
			try
			{
				Directory.CreateDirectory(this.currentImageDirectoryInfo.FullName);

				tileImageCreator.WorldDirectoryInfo = worldDirectoryInfo;
				tileImageCreator.ImageDirectoryInfo = this.currentImageDirectoryInfo;
				tileImageCreator.Biomes = this.biomes;
				tileImageCreator.BlendAll = this.blendAll;
				tileImageCreator.Direction = this.direction;
				tileImageCreator.Night = this.night;
				tileImageCreator.Skylight = this.skylight;
				tileImageCreator.WorldHeight = this.worldHeight;
				tileImageCreator.Hell = this.nether;
				tileImageCreator.Caves = this.caves;
				tileImageCreator.ChunkSize = 8;

				tileImageCreator.CreateTiles();
			}
			catch (Exception ex)
			{
				this.HandleException(ex);
			}
		}

		private void LoadTileIntoView(Tile tile)
		{
			double left = (tile.X * 32) - (tile.Z * -32);
			double top = (tile.Z * 16) + (tile.X * -16);

			tile.Left = left;
			tile.Top = top;
			tile.Bounds = new Rect(left, top, 522, 522);

			this.virtualCanvas.AddVirtualChild(tile);
		}

		private void HandleException(Exception ex)
		{
			if (!this.Dispatcher.CheckAccess())
			{
				this.Dispatcher.BeginInvoke(new Action<Exception>(this.HandleException), ex);
				return;
			}

			Debug.WriteLine(ex);
			logger.Error("", ex);
			MessageBox.Show(this, ex.ToString());
		}

		private void SaveAsImage()
		{
			SaveFileDialog dialog = new SaveFileDialog();
			dialog.Filter = "PNG Image (*.png)|*.png";
			var result = dialog.ShowDialog(this) ?? false;
			if (!result)
				return;

			this.SaveAsImage(dialog.FileName);
		}

		private void SaveAsImage(string path)
		{
			if (path == null) return;

			var surface = this.virtualCanvas.ContentCanvas;

			// Get the size of canvas
			Size size = new Size(this.virtualCanvas.ActualWidth, this.virtualCanvas.ActualHeight);

			// Create a render bitmap and push the surface to it
			RenderTargetBitmap renderBitmap =
			  new RenderTargetBitmap(
				 (int)size.Width,
				 (int)size.Height,
				 96d,
				 96d,
				 PixelFormats.Pbgra32);
			renderBitmap.Render(surface);

			// Create a file stream for saving image
			using (FileStream outStream = new FileStream(path, FileMode.Create))
			{
				// Use png encoder for our data
				PngBitmapEncoder encoder = new PngBitmapEncoder();
				// push the rendered bitmap to it
				encoder.Frames.Add(BitmapFrame.Create(renderBitmap));
				// save the data to the stream
				encoder.Save(outStream);
			}
		}

		private void ExportEntireImage()
		{
			SaveFileDialog dialog = new SaveFileDialog();
			dialog.Filter = "PNG Image (*.png)|*.png";
			var result = dialog.ShowDialog(this) ?? false;
			if (!result)
				return;

			this.ExportEntireImage(dialog.FileName);
		}

		private void ExportEntireImage(string imageFileName)
		{
			TileImageCreator tileImageCreator = new TileImageCreator(this.currentWorldDirectoryInfo, null);
			tileImageCreator.Biomes = this.biomes;
			tileImageCreator.BlendAll = this.blendAll;
			tileImageCreator.Direction = this.direction;
			tileImageCreator.Night = this.night;
			tileImageCreator.Skylight = this.skylight;
			tileImageCreator.ChunkSize = 8;
			tileImageCreator.WorldHeight = this.worldHeight;

			tileImageCreator.ExportEntireWorld(imageFileName);
		}

		private void DecreaseHeight()
		{
			try
			{
				int height = this.worldHeight - 1;
				SetWorldHeight(height);
			}
			catch (Exception ex)
			{
				this.HandleException(ex);
			}
		}

		private void IncreaseHeight()
		{
			try
			{
				int height = this.worldHeight + 1;
				SetWorldHeight(height);
			}
			catch (Exception ex)
			{
				this.HandleException(ex);
			}
		}

		private void SetWorldHeight(string heightString)
		{
			int height = 0;

			if (!int.TryParse(heightString, out height) || height < 1 || height > 128)
			{
				MessageBox.Show("Please enter a number between 1 and 128");
				return;
			}

			this.SetWorldHeight(height);
		}

		private void SetWorldHeight(int height)
		{
			Debug.WriteLine("New world height: " + height);

			if (this.worldHeight == height)
				return;

			if (height < 1 || height > 128)
				return;

			this.stackPanelHeight.IsEnabled = false;

			try
			{
				this.worldHeight = height;

				this.sliderHeight.ToolTip = "Height: " + height;
				this.textBoxHeight.Text = height.ToString();

				tileImageCreator.WorldDirectoryInfo = this.currentWorldDirectoryInfo;
				tileImageCreator.ImageDirectoryInfo = this.currentImageDirectoryInfo;
				tileImageCreator.Biomes = this.biomes;
				tileImageCreator.BlendAll = this.blendAll;
				tileImageCreator.Direction = this.direction;
				tileImageCreator.Night = this.night;
				tileImageCreator.Skylight = this.skylight;
				tileImageCreator.WorldHeight = this.worldHeight;
				tileImageCreator.ChunkSize = 8;

				foreach (var child in this.virtualCanvas.VirtualChildren)
				{
					Tile tile = child as Tile;

					tile.WorldHeight = height;

					string filename = string.Format("{0}.{1}.{2}.png", tile.X, tile.Z, height);

					filename = Path.Combine(this.currentImageDirectoryInfo.FullName, filename);

					tile.FileInfo = new FileInfo(filename);

					double top = (tile.Z * 16) + (tile.X * -16);

					top += (128 - height) * 2;

					tile.Top = top;
				}

				foreach (UIElement element in this.virtualCanvas.ContentCanvas.Children)
				{
					Image image = element as Image;
					if (image == null)
						continue;

					Tile tile = image.Tag as Tile;
					if (tile == null)
						continue;

					tile.FileInfo.Refresh();

					if (tile.FileInfo.Exists)
					{
						try
						{
							BitmapImage bitmapImage = new BitmapImage();
							bitmapImage.BeginInit();
							bitmapImage.CacheOption = BitmapCacheOption.OnDemand;
							bitmapImage.UriSource = new Uri(tile.FileInfo.FullName);
							bitmapImage.EndInit();

							tile.Image.Source = bitmapImage;
							tile.Image.SetValue(Canvas.TopProperty, tile.Top);
						}
						catch (Exception ex)
						{
							Debug.WriteLine(ex);
						}
					}
					else
					{
						tile.GenerateTileImageCallback(tile);
						//tileImageCreator.GenerateTileImage(tile);
					}
				}
			}
			finally
			{
				this.stackPanelHeight.IsEnabled = true;
			}
		}

		#endregion

		#region Event Handlers

		void MainWindow_Loaded(object sender, RoutedEventArgs e)
		{
			try
			{
				//Canvas target = virtualCanvas.ContentCanvas;

				//target.MouseMove += new MouseEventHandler(target_MouseMove);

				virtualCanvas.VisualsChanged += new EventHandler<VisualChangeEventArgs>(OnVisualsChanged);

				applicationData = Environment.GetFolderPath(System.Environment.SpecialFolder.ApplicationData);

				this.LoadMinecraftWorldMenuItems();

				this.ignoreDirectionCheck = false;

				this.menuItemBiomes.IsChecked = JavaChecker.EnsureJava(new Version(1, 5));
			}
			catch (Exception ex)
			{
				this.HandleException(ex);
			}
		}

		//Image hitImage;
		//void target_MouseMove(object sender, MouseEventArgs e)
		//{
		//   try
		//   {
		//      //if (hitImage != null)
		//      //   hitImage.Opacity = 1;

		//      if (this.selection != null)
		//         this.selection.Visibility = Visibility.Hidden;

		//      Canvas canvas = sender as Canvas;
		//      if (canvas == null)
		//         return;

		//      var hitPoint = e.GetPosition((UIElement)sender);

		//      hitImage = null;

		//      var result = VisualTreeHelper.HitTest(canvas, hitPoint);

		//      if (result == null)
		//         return;

		//      hitImage = result.VisualHit as Image;

		//      if (hitImage == null)
		//         return;

		//      //hitImage.Opacity = .5;

		//      this.selection.SetValue(Canvas.LeftProperty, hitImage.GetValue(Canvas.LeftProperty));
		//      this.selection.SetValue(Canvas.TopProperty, hitImage.GetValue(Canvas.TopProperty));
		//      int zindex = int.MaxValue;
		//      this.selection.SetValue(Canvas.ZIndexProperty, zindex);
		//      this.selection.Visibility = Visibility.Visible;

		//      if (!canvas.Children.Contains(this.selection))
		//         canvas.Children.Add(this.selection);
		//   }
		//   catch (Exception ex)
		//   {
		//      this.HandleException(ex);
		//   }
		//}

		void OnVisualsChanged(object sender, VisualChangeEventArgs e)
		{
			try
			{
				if (this.virtualCanvas.VirtualChildren == null || this.virtualCanvas.VirtualChildren.Count < 1)
				{
					//StatusText.Text = "";
				}
				else
				{
					long current = this.virtualCanvas.LiveVisualCount * 64;
					long total = this.virtualCanvas.VirtualChildren.Count * 64;
					StatusText.Text = string.Format("Showing {0:N0} of {1:N0} chunks", current, total);
				}

				int tick = Environment.TickCount;
				if (e.Added != 0 || e.Removed != 0)
				{
					addedPerSecond += e.Added;
					removedPerSecond += e.Removed;
					if (tick > lastTick + 100)
					{
						Created.BeginAnimation(System.Windows.Shapes.Rectangle.WidthProperty, new DoubleAnimation(
							 Math.Min(addedPerSecond, 450),
							 new Duration(TimeSpan.FromMilliseconds(100))));
						CreatedLabel.Text = addedPerSecond.ToString() + " tiles created";
						addedPerSecond = 0;

						Destroyed.BeginAnimation(System.Windows.Shapes.Rectangle.WidthProperty, new DoubleAnimation(
							 Math.Min(removedPerSecond, 450),
							 new Duration(TimeSpan.FromMilliseconds(100))));
						DestroyedLabel.Text = removedPerSecond.ToString() + " tiles disposed";
						removedPerSecond = 0;
					}
				}
				if (tick > lastTick + 1000)
				{
					lastTick = tick;
				}
			}
			catch (Exception ex)
			{
				this.HandleException(ex);
			}
		}

		void menuItemBrowseForWorld_Click(object sender, RoutedEventArgs e)
		{
			try
			{
				using (System.Windows.Forms.FolderBrowserDialog dialog = new System.Windows.Forms.FolderBrowserDialog())
				{
					dialog.SelectedPath = Path.Combine(applicationData, ".minecraft", "saves");
					if (dialog.ShowDialog() != System.Windows.Forms.DialogResult.OK)
						return;

					this.LoadWorld(new DirectoryInfo(dialog.SelectedPath));
				}
			}
			catch (Exception ex)
			{
				this.HandleException(ex);
			}
		}

		void menuItemRefresh_Click(object sender, RoutedEventArgs e)
		{
			try
			{
				this.LoadMinecraftWorldMenuItems();
				this.ReloadWorld();
			}
			catch (Exception ex)
			{
				this.HandleException(ex);
			}
		}

		void menuItemViewCacheFolder_Click(object sender, RoutedEventArgs e)
		{
			try
			{				
				string path = Path.Combine(Environment.GetFolderPath(Environment.SpecialFolder.ApplicationData), "Minemapper");

				Process.Start(path);
			}
			catch (Exception ex)
			{
				this.HandleException(ex);
			}
		}

		void OnLoadWorldMenuItem_Checked(object sender, RoutedEventArgs e)
		{
			try
			{
				if (ignoreWorldCheck)
					return;

				ignoreWorldCheck = true;

				MenuItem menuItem = sender as MenuItem;
				if (menuItem == null)
					return;

				DirectoryInfo directoryInfo = menuItem.Tag as DirectoryInfo;
				if (directoryInfo == null)
					return;

				this.LoadWorld(directoryInfo);
			}
			catch (Exception ex)
			{
				this.HandleException(ex);
			}
			finally
			{
				ignoreWorldCheck = false;
			}
		}

		void tileImageCreator_TilesCreated(object sender, EventArgs<List<Tile>> e)
		{
			try
			{
				if (!this.Dispatcher.CheckAccess())
				{
					this.Dispatcher.BeginInvoke(new EventHandler<EventArgs<List<Tile>>>(this.tileImageCreator_TilesCreated), DispatcherPriority.ContextIdle, sender, e);
					return;
				}

				this.SetWorldControlsState(true);

				var tiles = e.Item;

				if (tiles == null || tiles.Count < 1)
				{
					this.StatusText.Text = "The selected world is empty.";
					return;
				}

				this.StatusBar.Visibility = this.menuItemStatusBar.IsChecked ? Visibility.Visible : Visibility.Collapsed;
				
				foreach (Tile tile in tiles)
				{
					this.LoadTileIntoView(tile);
				}
			}
			catch (Exception ex)
			{
				this.HandleException(ex);
			}
		}

		void tileImageCreator_TileImageLoaded(object sender, EventArgs<Tile> e)
		{
			try
			{
				if (!this.Dispatcher.CheckAccess())
				{
					this.Dispatcher.BeginInvoke(new EventHandler<EventArgs<Tile>>(this.tileImageCreator_TileImageLoaded), DispatcherPriority.ContextIdle, sender, e);
					return;
				}

				Tile tile = e.Item;

				if (tile == null)
					return;

				if (tile.FileInfo == null)
					return;

				tile.FileInfo.Refresh();

				if (tile.Image == null)
					return;

				tile.Image.Tag = tile;

				if (tile.FileInfo.Exists)
				{
					// is the file blank?
					if (tile.FileInfo.Length <= 1163)
					{
						//// this should improve performace, by removing unneeded virtual tiles,
						//// but it actually forces a complete recalculation every time it 
						//// removes a tile.  Gotta find a better way.
						//this.virtualCanvas.VirtualChildren.Remove(tile);

						tile.IsBlank = true;
						tile.Image.Source = null;
					}
					else
					{
						RenderOptions.SetBitmapScalingMode(tile.Image, BitmapScalingMode.NearestNeighbor);
						RenderOptions.SetEdgeMode(tile.Image, EdgeMode.Aliased);

						BitmapImage bitmapImage = new BitmapImage();
						bitmapImage.BeginInit();
						bitmapImage.CacheOption = BitmapCacheOption.OnDemand;
						bitmapImage.UriSource = new Uri(tile.FileInfo.FullName);
						bitmapImage.EndInit();

						tile.Image.Source = bitmapImage;
						tile.Image.SetValue(Canvas.TopProperty, tile.Top);
					}
				}
				else
				{
					RenderOptions.SetBitmapScalingMode(tile.Image, BitmapScalingMode.HighQuality);
					RenderOptions.SetEdgeMode(tile.Image, EdgeMode.Unspecified);

					if (Tile.EmptyImageUri == null)
					{
						string path = Path.GetDirectoryName(System.Reflection.Assembly.GetExecutingAssembly().Location);
						path = Path.Combine(path, "empty.png");
						Tile.EmptyImageUri = new Uri(path);
					}

					tile.Image.Source = new BitmapImage(Tile.EmptyImageUri);
				}
			}
			catch (Exception ex)
			{
				Debug.WriteLine(ex);
				logger.Error("", ex);
			}
		}

		void tileImageCreator_StatusChanged(object sender, EventArgs<string> e)
		{
			try
			{
				if (!this.Dispatcher.CheckAccess())
				{
					this.Dispatcher.BeginInvoke(new EventHandler<EventArgs<string>>(this.tileImageCreator_StatusChanged), DispatcherPriority.ContextIdle, sender, e);
					return;
				}

				this.StatusText.Text = e.Item;
			}
			catch (Exception ex)
			{
				Debug.WriteLine(ex);
				logger.Error("", ex);
			}
		}

		void tileImageCreator_ProgressChanged(object sender, EventArgs<string> e)
		{
			try
			{
				if (!this.Dispatcher.CheckAccess())
				{
					this.Dispatcher.BeginInvoke(new EventHandler<EventArgs<string>>(this.tileImageCreator_ProgressChanged), DispatcherPriority.ContextIdle, sender, e);
					return;
				}

				this.ProgressText.Text = e.Item;
			}
			catch (Exception ex)
			{
				Debug.WriteLine(ex);
				logger.Error("", ex);
			}
		}

		void tileImageCreator_ProcessCountChanged(object sender, EventArgs<int> e)
		{
			try
			{
				if (!this.Dispatcher.CheckAccess())
				{
					this.Dispatcher.BeginInvoke(new EventHandler<EventArgs<int>>(this.tileImageCreator_ProcessCountChanged), DispatcherPriority.ContextIdle, sender, e);
					return;
				}

				this.ProcessCountText.Text = string.Format("Processes: {0}", e.Item);
			}
			catch (Exception ex)
			{
				Debug.WriteLine(ex);
				logger.Error("", ex);
			}
		}

		private void menuItemDirection_Checked(object sender, RoutedEventArgs e)
		{
			try
			{
				if (this.ignoreDirectionCheck)
					return;

				MenuItem menuItem = sender as MenuItem;
				if (menuItem == null)
					return;

				string header = menuItem.Header as string;
				if (string.IsNullOrEmpty(header))
					return;

				ignoreDirectionCheck = true;

				try
				{
					this.menuItemNorth.IsChecked = menuItem == this.menuItemNorth;
					this.menuItemSouth.IsChecked = menuItem == this.menuItemSouth;
					this.menuItemEast.IsChecked = menuItem == this.menuItemEast;
					this.menuItemWest.IsChecked = menuItem == this.menuItemWest;
				}
				finally
				{
					ignoreDirectionCheck = false;
				}

				direction = header.ToLower();

				this.ReloadWorld();
			}
			catch (Exception ex)
			{
				this.HandleException(ex);
			}
		}

		private void menuItemBiomes_Checked(object sender, RoutedEventArgs e)
		{
			if (!this.IsLoaded)
				return;

			if (this.ignoreModeCheck)
				return;

			try
			{
				this.ignoreModeCheck = true;

				if (this.menuItemBiomes.IsChecked)
				{
					if (!JavaChecker.EnsureJava(new Version(1, 5)))
					{
						this.menuItemBiomes.IsChecked = false;
						MessageBox.Show("Java is required for displaying biomes.  Minemapper could not find Java in the 'path' Windows environment variable or registry.");
						return;
					}
				}

				if (this.biomes == this.menuItemBiomes.IsChecked)
					return;

				this.biomes = this.menuItemBiomes.IsChecked;
				this.nether = false;

				this.ReloadWorld();
			}
			catch (Exception ex)
			{
				this.HandleException(ex);
			}
			finally
			{
				this.ignoreModeCheck = false;
			}
		}

		private void OnModeChanged(object sender, SelectionChangedEventArgs e)
		{
			try
			{
				if (!this.IsLoaded)
					return;

				this.ignoreModeCheck = true;

				string mode = this.comboBoxMode.SelectedItem as string;

				this.night = false;
				this.nether = false;
				this.caves = false;

				this.menuItemBiomes.IsEnabled = true;

				switch (mode)
				{
					case "Day":
						break;

					case "Night":
						this.night = true;
						break;

					case "Caves":
						this.caves = true;
						break;

					case "Nether":
						this.nether = true;
						this.biomes = false;
						this.menuItemBiomes.IsChecked = false;
						this.menuItemBiomes.IsEnabled = false;
						break;
				}

				this.ReloadWorld();
			}
			catch (Exception ex)
			{
				this.HandleException(ex);
			}
			finally
			{
				this.ignoreModeCheck = false;
			}
		}

		private void menuItemSkylight_Checked(object sender, RoutedEventArgs e)
		{
			try
			{
				this.skylight = this.menuItemSkylight.IsChecked;

				this.ReloadWorld();
			}
			catch (Exception ex)
			{
				this.HandleException(ex);
			}
		}

		private void menuItemBlendAll_Checked(object sender, RoutedEventArgs e)
		{
			try
			{
				this.blendAll = this.menuItemBlendAll.IsChecked;

				this.ReloadWorld();
			}
			catch (Exception ex)
			{
				this.HandleException(ex);
			}
		}

		private void sliderZoom_ValueChanged(object sender, RoutedPropertyChangedEventArgs<double> e)
		{
			try
			{
				ignoreScaleChanges = true;

				double zoom = e.NewValue;

				this.textBlockZoom.Text = zoom.ToString("P0");
				this.sliderZoom.ToolTip = this.textBlockZoom.Text;

				if (ignoreSliderValueChanges)
					return;

				this.virtualCanvas.SetZoom(zoom);
			}
			catch (Exception ex)
			{
				this.HandleException(ex);
			}
			finally
			{
				ignoreScaleChanges = false;
			}
		}

		void ScaleTransform_Changed(object sender, EventArgs e)
		{
			try
			{
				if (ignoreScaleChanges)
					return;

				ignoreSliderValueChanges = true;

				double value = this.virtualCanvas.ScaleTransform.ScaleX;

				if (value < this.sliderZoom.Minimum)
					value = this.sliderZoom.Minimum;

				if (value > this.sliderZoom.Maximum)
					value = this.sliderZoom.Maximum;

				this.sliderZoom.Value = value;
			}
			catch (Exception ex)
			{
				this.HandleException(ex);
			}
			finally
			{
				ignoreSliderValueChanges = false;
			}
		}

		private void menuItemSaveImage_Click(object sender, RoutedEventArgs e)
		{
			try
			{
				this.SaveAsImage();
			}
			catch (Exception ex)
			{
				this.HandleException(ex);
			}
		}

		private void menuItemSaveEntireWorldImage_Click(object sender, RoutedEventArgs e)
		{
			try
			{
				this.ExportEntireImage();
			}
			catch (Exception ex)
			{
				this.HandleException(ex);
			}
		}

		private void OnDecreaseHeight(object sender, EventArgs e)
		{
			try
			{
				this.DecreaseHeight();
			}
			catch (Exception ex)
			{
				this.HandleException(ex);
			}
		}

		private void OnIncreaseHeight(object sender, EventArgs e)
		{
			try
			{
				this.IncreaseHeight();
			}
			catch (Exception ex)
			{
				this.HandleException(ex);
			}
		}

		private void sliderHeight_ValueChanged(object sender, RoutedPropertyChangedEventArgs<double> e)
		{
			try
			{
				int height = (int)e.NewValue;
				SetWorldHeight(height);
			}
			catch (Exception ex)
			{
				this.HandleException(ex);
			}
		}

		private void ZoomInCommandBinding_CanExecute(object sender, CanExecuteRoutedEventArgs e)
		{
			try
			{
				e.CanExecute = true;
			}
			catch (Exception ex)
			{
				this.HandleException(ex);
			}
		}

		private void ZoomInCommandBinding_Executed(object sender, ExecutedRoutedEventArgs e)
		{
			try
			{
				double value = this.sliderZoom.Value;

				value += .1;

				if (value > this.sliderZoom.Maximum)
					value = this.sliderZoom.Maximum;

				this.sliderZoom.Value = value;
			}
			catch (Exception ex)
			{
				this.HandleException(ex);
			}
		}

		private void ZoomOutCommandBinding_CanExecute(object sender, CanExecuteRoutedEventArgs e)
		{
			try
			{
				e.CanExecute = true;
			}
			catch (Exception ex)
			{
				this.HandleException(ex);
			}
		}

		private void ZoomOutCommandBinding_Executed(object sender, ExecutedRoutedEventArgs e)
		{
			try
			{
				double value = this.sliderZoom.Value;

				value -= .1;

				if (value < this.sliderZoom.Minimum)
					value = this.sliderZoom.Minimum;

				this.sliderZoom.Value = value;
			}
			catch (Exception ex)
			{
				this.HandleException(ex);
			}
		}

		private void PanUpCommandBinding_CanExecute(object sender, CanExecuteRoutedEventArgs e)
		{
			try
			{
				e.CanExecute = true;
			}
			catch (Exception ex)
			{
				this.HandleException(ex);
			}
		}

		private void PanUpCommandBinding_Executed(object sender, ExecutedRoutedEventArgs e)
		{
			try
			{
				this.virtualCanvas.PanUp();
			}
			catch (Exception ex)
			{
				this.HandleException(ex);
			}
		}

		private void PanDownCommandBinding_CanExecute(object sender, CanExecuteRoutedEventArgs e)
		{
			try
			{
				e.CanExecute = true;
			}
			catch (Exception ex)
			{
				this.HandleException(ex);
			}
		}

		private void PanDownCommandBinding_Executed(object sender, ExecutedRoutedEventArgs e)
		{
			try
			{
				this.virtualCanvas.PanDown();
			}
			catch (Exception ex)
			{
				this.HandleException(ex);
			}
		}

		private void PanLeftCommandBinding_CanExecute(object sender, CanExecuteRoutedEventArgs e)
		{
			try
			{
				e.CanExecute = true;
			}
			catch (Exception ex)
			{
				this.HandleException(ex);
			}
		}

		private void PanLeftCommandBinding_Executed(object sender, ExecutedRoutedEventArgs e)
		{
			try
			{
				this.virtualCanvas.PanLeft();
			}
			catch (Exception ex)
			{
				this.HandleException(ex);
			}
		}

		private void PanRightCommandBinding_CanExecute(object sender, CanExecuteRoutedEventArgs e)
		{
			try
			{
				e.CanExecute = true;
			}
			catch (Exception ex)
			{
				this.HandleException(ex);
			}
		}

		private void PanRightCommandBinding_Executed(object sender, ExecutedRoutedEventArgs e)
		{
			try
			{
				this.virtualCanvas.PanRight();
			}
			catch (Exception ex)
			{
				this.HandleException(ex);
			}
		}

		private void ResetViewCommandBinding_CanExecute(object sender, CanExecuteRoutedEventArgs e)
		{
			try
			{
				e.CanExecute = true;
			}
			catch (Exception ex)
			{
				this.HandleException(ex);
			}
		}

		private void ResetViewCommandBinding_Executed(object sender, ExecutedRoutedEventArgs e)
		{
			try
			{
				this.virtualCanvas.Reset();
			}
			catch (Exception ex)
			{
				this.HandleException(ex);
			}
		}

		private void OnChangeBackgroundColor(object sender, RoutedEventArgs e)
		{
			try
			{
				using (System.Windows.Forms.ColorDialog dialog = new System.Windows.Forms.ColorDialog())
				{
					if (dialog.ShowDialog() != System.Windows.Forms.DialogResult.OK)
						return;

					var color = dialog.Color;

					var wpfColor = Color.FromArgb(color.A, color.R, color.G, color.B);

					var brush = new SolidColorBrush(wpfColor);

					this.virtualCanvas.Background = brush;
					this.virtualCanvas.ContentCanvas.Background = brush;
					this.virtualCanvas.Backdrop.Background = brush;

					//if (wpfColor == Colors.White)
					//   this.textBlockZoom.Background = Brushes.Transparent;
					//else
					//   this.textBlockZoom.Background = Brushes.White;
				}
			}
			catch (Exception ex)
			{
				this.HandleException(ex);
			}
		}

		private void OnSetHeight(object sender, RoutedEventArgs e)
		{
			try
			{
				int height = 0;

				if (!int.TryParse(this.textBoxHeight.Text, out height) || height < 1 || height > 128)
				{
					MessageBox.Show("Please enter a number between 1 and 128");
					return;
				}

				this.SetWorldHeight(height);
			}
			catch (Exception ex)
			{
				this.HandleException(ex);
			}
		}

		private void textBoxHeight_KeyDown(object sender, KeyEventArgs e)
		{
			try
			{
				if (e.Key != Key.Enter)
					return;

				string heightString = this.textBoxHeight.Text;

				this.SetWorldHeight(heightString);
			}
			catch (Exception ex)
			{
				this.HandleException(ex);
			}
		}

		private void menuItemStatusBar_Checked(object sender, RoutedEventArgs e)
		{
			try
			{
				if (!this.IsLoaded)
					return;

				this.StatusBar.Visibility = this.menuItemStatusBar.IsChecked ? Visibility.Visible : Visibility.Collapsed;
			}
			catch (Exception ex)
			{
				this.HandleException(ex);
			}
		}

		#endregion
	}
}
