﻿using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.IO;
using System.Text.RegularExpressions;
using System.Threading.Tasks;

namespace Minemapper
{
	public class TileImageCreator
	{
		public event EventHandler<EventArgs<Tile>> TileImageLoaded;
		public event EventHandler<EventArgs<List<Tile>>> TilesCreated;
		public event EventHandler<EventArgs<String>> StatusChanged;
		public event EventHandler<EventArgs<String>> ProgressChanged;
		public event EventHandler<EventArgs<int>> ProcessCountChanged;

		public DirectoryInfo WorldDirectoryInfo { get; set; }
		public DirectoryInfo ImageDirectoryInfo { get; set; }

		public string Direction { get; set; }
		public bool Biomes { get; set; }
		public bool Night { get; set; }
		public bool Skylight { get; set; }
		public bool BlendAll { get; set; }
		public bool Hell { get; set; }
		public bool Caves { get; set; }

		public int WorldHeight { get; set; }

		public int ChunkSize { get; set; }

		private int processCount;
		public int ProcessCount
		{
			get
			{
				return processCount;
			}

			set
			{
				processCount = value;

				if (this.ProcessCountChanged != null)
					this.ProcessCountChanged(this, new EventArgs<int>(processCount));
			}
		}

		public TileImageCreator(DirectoryInfo worldDirectoryInfo, DirectoryInfo imageDirectoryInfo)
		{
			this.WorldDirectoryInfo = worldDirectoryInfo;
			this.ImageDirectoryInfo = imageDirectoryInfo;
		}

		public TileImageCreator()
		{
			// TODO: Complete member initialization
		}

		public void CreateTiles()
		{
			List<Tile> tiles = new List<Tile>();

			try
			{
				if (!ImageDirectoryInfo.Exists)
					Directory.CreateDirectory(ImageDirectoryInfo.FullName);

				if (this.StatusChanged != null)
					this.StatusChanged(this, new EventArgs<string>("Calculating world size..."));

				WorldBounds worldBounds = this.GetWorldSize();

				if (worldBounds == null)
					return;

				if (this.Biomes)
				{
					if (this.StatusChanged != null)
						this.StatusChanged(this, new EventArgs<string>("Extracting/updating biome information..."));

					this.ExtractBiomes();
				}

				if (this.StatusChanged != null)
					this.StatusChanged(this, new EventArgs<string>("Adding tiles..."));

				switch (this.Direction.ToLower())
				{
					case "west":
						for (long x = worldBounds.MaxX; x <= worldBounds.MinX; x += this.ChunkSize)
						{
							for (long z = worldBounds.MaxZ; z <= worldBounds.MinZ; z += this.ChunkSize)
							{
								AddTile(-x, -z, tiles);
							}
						}
						break;

					case "north":
						for (long x = worldBounds.MaxZ; x >= worldBounds.MinZ; x -= this.ChunkSize)
						{
							for (long z = worldBounds.MinX; z >= worldBounds.MaxX; z -= this.ChunkSize)
							{
								AddTile(-x, z, tiles);
							}
						}
						break;

					case "east":
					default:
						for (long x = worldBounds.MinX; x <= worldBounds.MaxX; x += this.ChunkSize)
						{
							for (long z = worldBounds.MinZ; z <= worldBounds.MaxZ; z += this.ChunkSize)
							{
								AddTile(x, z, tiles);
							}
						}
						break;
				}
			}
			finally
			{
				if (this.TilesCreated != null)
					this.TilesCreated(this, new EventArgs<List<Tile>>(tiles));
			}
		}

		private void AddTile(long x, long z, List<Tile> tiles)
		{
			try
			{
				string filename = string.Format("{0}.{1}.{2}.png", x, z, this.WorldHeight);

				filename = Path.Combine(ImageDirectoryInfo.FullName, filename);

				Tile tile = new Tile() { X = x, Z = z, FileInfo = new FileInfo(filename) };

				tile.WorldHeight = this.WorldHeight;

				tile.GenerateTileImageCallback = this.GenerateTileImageAsync;

				tiles.Add(tile);
			}
			catch (Exception ex)
			{
				Debug.WriteLine(ex);
			}
		}

		public void GenerateTileImageAsync(Tile tile)
		{
			Action<object> action = new Action<object>(this.GenerateTileImage);

			TaskCreationOptions options = TaskCreationOptions.PreferFairness;

			var task = Task.Factory.StartNew(action, tile, options);
		}

		private void GenerateTileImage(object state)
		{
			this.GenerateTileImage(state as Tile);
		}

		public void GenerateTileImage(Tile tile)
		{
			double x = tile.X;
			double z = tile.Z;

			switch (this.Direction.ToLower())
			{
				case "west":
					x = -x;
					z = -z;
					break;

				//case "north":
				//   z = -x;
				//   x = z;
				//   break;
			}

			string filename = string.Format("{0}.{1}.{2}.png", tile.X, tile.Z, this.WorldHeight);

			filename = Path.Combine(this.ImageDirectoryInfo.FullName, filename);

			tile.WorldHeight = this.WorldHeight;
			tile.FileInfo = new FileInfo(filename);

			string arguments = string.Format("-from {0} {1} -to {2} {3} -png -file \"{4}\" \"{5}\"", x, z, x + 7, z + 7, tile.FileInfo.FullName, WorldDirectoryInfo.FullName);

			if (this.Skylight)
				arguments += " -skylight";

			if (!string.IsNullOrEmpty(this.Direction))
				arguments += " -" + this.Direction;

			if (this.Night)
				arguments += " -night";
			else if (this.Hell)
				arguments += " -nether";
			else if (this.Caves)
				arguments += " -cave";

			if (this.BlendAll)
				arguments += " -blendall";

			if (this.Biomes && !this.Hell)
				arguments += " -biomes";

			arguments += " -height " + this.WorldHeight;

			string mcmap = Environment.Is64BitOperatingSystem ? "mcmap64.exe" : "mcmap32.exe";

			ProcessStartInfo info = new ProcessStartInfo(mcmap, arguments);
			info.CreateNoWindow = true;
			info.UseShellExecute = false;

			Process process = new Process();
			process.StartInfo = info;
			process.EnableRaisingEvents = true;
			process.Exited += delegate(object sender, EventArgs e)
			{
				this.ProcessCount--;

				tile.FileInfo.Refresh();

				if (!tile.FileInfo.Exists)
					return;

				if (this.TileImageLoaded != null)
					this.TileImageLoaded(this, new EventArgs<Tile>(tile));
			};

			Debug.WriteLine("Generating image: " + tile.FileInfo.Name);

			process.Start();
			this.ProcessCount++;

			process.WaitForExit();
		}

		/// <summary>
		///  Regular expression built for C# on: Thu, Nov 4, 2010, 01:59:30 PM
		///  Using Expresso Version: 3.0.3634, http://www.ultrapico.com
		///  
		///  A description of the regular expression:
		///  
		///  \s*Min:\s*\(\s*
		///      Whitespace, any number of repetitions
		///      Min:
		///      Whitespace, any number of repetitions
		///      Literal (
		///      Whitespace, any number of repetitions
		///  [MinX]: A named capture group. [-?\d+]
		///      -?\d+
		///          -, zero or one repetitions
		///          Any digit, one or more repetitions
		///  \s*\|\s*
		///      Whitespace, any number of repetitions
		///      Literal |
		///      Whitespace, any number of repetitions
		///  [MinZ]: A named capture group. [-?\d+]
		///      -?\d+
		///          -, zero or one repetitions
		///          Any digit, one or more repetitions
		///  \s*\)\s*Max:\s*\(\s*
		///      Whitespace, any number of repetitions
		///      Literal )
		///      Whitespace, any number of repetitions
		///      Max:
		///      Whitespace, any number of repetitions
		///      Literal (
		///      Whitespace, any number of repetitions
		///  [MaxX]: A named capture group. [-?\d+]
		///      -?\d+
		///          -, zero or one repetitions
		///          Any digit, one or more repetitions
		///  \s*\|\s*
		///      Whitespace, any number of repetitions
		///      Literal |
		///      Whitespace, any number of repetitions
		///  [MaxZ]: A named capture group. [-?\d+]
		///      -?\d+
		///          -, zero or one repetitions
		///          Any digit, one or more repetitions
		///  \s*\)\s*
		///      Whitespace, any number of repetitions
		///      Literal )
		///      Whitespace, any number of repetitions
		///  
		///
		/// </summary>
		public static Regex regex = new Regex(
				"\\s*Min:\\s*\\(\\s*(?<MinX>-?\\d+)\\s*\\|\\s*(?<MinZ>-?\\d+)" +
				"\\s*\\)\\s*Max:\\s*\\(\\s*(?<MaxX>-?\\d+)\\s*\\|\\s*(?<MaxZ>" +
				"-?\\d+)\\s*\\)\\s*",
			 RegexOptions.IgnoreCase
			 | RegexOptions.CultureInvariant
			 | RegexOptions.Compiled
			 );

		private void ExtractBiomes()
		{
			string args = string.Format("-jar MinecraftBiomeExtractor.jar -nogui \"{0}\"", WorldDirectoryInfo.FullName);

			ProcessStartInfo info = new ProcessStartInfo("java.exe", args);
			info.UseShellExecute = false;
			info.CreateNoWindow = true;
			info.WindowStyle = ProcessWindowStyle.Hidden;

			Process process = new Process();
			process.StartInfo = info;
			process.Start();
			this.ProcessCount++;
			process.WaitForExit();
			this.ProcessCount--;
		}

		private WorldBounds GetWorldSize()
		{
			string filename = Path.Combine(ImageDirectoryInfo.FullName, WorldDirectoryInfo.Name + ".png");

			string arguments = string.Format("-png -file \"{0}\" \"{1}\"", filename, WorldDirectoryInfo.FullName);

			if (!string.IsNullOrEmpty(this.Direction))
				arguments += " -" + this.Direction;

			if (this.Hell)
				arguments += " -nether";

			if (this.Caves)
				arguments += " -cave";

			string mcmap = Environment.Is64BitOperatingSystem ? "mcmap64.exe" : "mcmap32.exe";

			ProcessStartInfo info = new ProcessStartInfo(mcmap, arguments);
			info.UseShellExecute = false;
			info.CreateNoWindow = true;
			info.WindowStyle = ProcessWindowStyle.Hidden;
			info.RedirectStandardOutput = true;

			Process process = new Process();
			process.StartInfo = info;
			process.EnableRaisingEvents = true;

			WorldBounds worldBounds = null;

			process.OutputDataReceived += delegate(object sender, DataReceivedEventArgs e)
			{
				Debug.WriteLine(e.Data);

				if (this.ProgressChanged != null)
					this.ProgressChanged(this, new EventArgs<string>(e.Data));

				if (string.IsNullOrEmpty(e.Data))
					return;

				// look for something like:
				// Min: (-81|-23) Max: (312|63)

				if (!regex.IsMatch(e.Data))
					return;

				var match = regex.Match(e.Data);
				string minXText = match.Groups["MinX"].Value;
				string minZText = match.Groups["MinZ"].Value;
				string maxXText = match.Groups["MaxX"].Value;
				string maxZText = match.Groups["MaxZ"].Value;

				worldBounds = WorldBounds.Parse(minXText, minZText, maxXText, maxZText);

				worldBounds.RoundUp(this.ChunkSize);

				long maxX;
				long maxZ;

				switch (this.Direction.ToLower())
				{
					case "west":
						maxX = worldBounds.MaxX;
						worldBounds.MaxX = worldBounds.MinX;
						worldBounds.MinX = maxX;

						maxZ = worldBounds.MaxZ;
						worldBounds.MaxZ = worldBounds.MinZ;
						worldBounds.MinZ = maxZ;
						break;

					case "north":
						WorldBounds rotatedWorldBounds = new WorldBounds()
						{
							MinX = worldBounds.MaxZ,
							MaxX = worldBounds.MinZ,
							MinZ = worldBounds.MinX,
							MaxZ = worldBounds.MaxX,
						};
						worldBounds = rotatedWorldBounds;
						break;
				}

				process.CancelOutputRead();
				if (!process.HasExited)
					try { process.Kill(); }
					catch (Exception ex) { Debug.WriteLine(ex); }
			};

			process.Start();
			this.ProcessCount++;
			process.BeginOutputReadLine();
			process.WaitForExit();
			this.ProcessCount--;

			return worldBounds;
		}

		public void ExportEntireWorld(string imageFileName)
		{
			string arguments = string.Format("-png -file \"{0}\" \"{1}\"", imageFileName, WorldDirectoryInfo.FullName);

			if (this.Skylight)
				arguments += " -skylight";

			if (!string.IsNullOrEmpty(this.Direction))
				arguments += " -" + this.Direction;

			if (this.Night)
				arguments += " -night";
			else if (this.Hell)
				arguments += " -nether";
			else if (this.Caves)
				arguments += " -cave";

			if (this.BlendAll)
				arguments += " -blendall";

			if (this.Biomes && !this.Hell)
				arguments += " -biomes";

			arguments += " -height " + this.WorldHeight;

			string mcmap = Environment.Is64BitOperatingSystem ? "mcmap64.exe" : "mcmap32.exe";

			ProcessStartInfo info = new ProcessStartInfo(mcmap, arguments);
			info.UseShellExecute = false;

			Process process = new Process();
			process.StartInfo = info;
			process.EnableRaisingEvents = true;
			process.Exited += delegate(object sender, EventArgs e)
			{
				this.ProcessCount--;
			};

			process.Start();
			this.processCount++;
		}
	}

	public class WorldBounds
	{
		public long MinX { get; set; }
		public long MinZ { get; set; }
		public long MaxX { get; set; }
		public long MaxZ { get; set; }

		public WorldBounds()
		{
		}

		public WorldBounds(long minX, long minZ, long MaxX, long MaxZ)
		{
			this.MinX = minX;
			this.MinZ = minZ;
			this.MaxX = MaxX;
			this.MaxZ = MaxZ;
		}

		public static WorldBounds Parse(string minXText, string minZText, string maxXText, string maxZText)
		{
			long minX = long.Parse(minXText);
			long minZ = long.Parse(minZText);
			long maxX = long.Parse(maxXText);
			long maxZ = long.Parse(maxZText);

			WorldBounds worldBounds = new WorldBounds(minX, minZ, maxX, maxZ);

			return worldBounds;
		}

		public void RoundUp(long chunkSize)
		{
			// round the bounds up to the nearest chunkSize
			long remainder = 0;

			long quotient = Math.DivRem(this.MinX, chunkSize, out remainder);
			if (remainder < 0)
				this.MinX -= chunkSize + remainder;
			else if (remainder > 0)
				this.MinX += chunkSize - remainder;

			quotient = Math.DivRem(this.MinZ, chunkSize, out remainder);
			if (remainder < 0)
				this.MinZ -= chunkSize + remainder;
			else if (remainder > 0)
				this.MinZ += chunkSize - remainder;

			quotient = Math.DivRem(this.MaxX, chunkSize, out remainder);
			if (remainder < 0)
				this.MaxX -= chunkSize + remainder;
			else if (remainder > 0)
				this.MaxX += chunkSize - remainder;

			quotient = Math.DivRem(this.MaxZ, chunkSize, out remainder);
			if (remainder < 0)
				this.MaxZ -= chunkSize + remainder;
			else if (remainder > 0)
				this.MaxZ += chunkSize - remainder;
		}
	}
}
