﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Windows.Controls;
using System.Windows;
using System.Windows.Media;
using System.Windows.Input;
using System.Windows.Media.Animation;
using System.Collections.ObjectModel;
using System.Collections.Specialized;
using System.Windows.Threading;
using System.Globalization;
using System.IO;
using System.Windows.Controls.Primitives;
using System.Diagnostics;
using log4net;

namespace Microsoft.Sample.Controls
{
	public class PanAndZoomVirtualCanvas : VirtualizingPanel, IScrollInfo
	{
		private static readonly ILog logger = LogManager.GetLogger(typeof(PanAndZoomVirtualCanvas));

		ScrollViewer _owner;
		Size _viewPortSize;
		bool _canHScroll = false;
		bool _canVScroll = false;
		QuadTree<IVirtualChild> _index;
		ObservableCollection<IVirtualChild> _children;
		Size _smallScrollIncrement = new Size(10, 10);
		public Canvas Content;
		Border _backdrop;
		Size _extent;
		IList<Rect> _dirtyRegions = new List<Rect>();
		IList<Rect> _visibleRegions = new List<Rect>();
		IDictionary<IVirtualChild, int> _visualPositions;
		int _nodeCollectCycle;
		bool _done = true;
		PathGeometry _clip;

		public double DefaultZoomFactor { get; set; }
		private Point ScreenStartPoint = new Point(0, 0);
		private TransformGroup transformGroup;
		private Point startOffset;

		public TranslateTransform TranslateTransform
		{
			get { return (TranslateTransform)GetValue(TranslateTransformProperty); }
			set { SetValue(TranslateTransformProperty, value); }
		}

		// Using a DependencyProperty as the backing store for TranslateTransform.  This enables animation, styling, binding, etc...
		public static readonly DependencyProperty TranslateTransformProperty =
			 DependencyProperty.Register("TranslateTransform", typeof(TranslateTransform), typeof(PanAndZoomVirtualCanvas), new UIPropertyMetadata(null));

		public ScaleTransform ScaleTransform
		{
			get { return (ScaleTransform)GetValue(ScaleTransformProperty); }
			set { SetValue(ScaleTransformProperty, value); }
		}

		// Using a DependencyProperty as the backing store for ScaleTransform.  This enables animation, styling, binding, etc...
		public static readonly DependencyProperty ScaleTransformProperty =
			 DependencyProperty.Register("ScaleTransform", typeof(ScaleTransform), typeof(PanAndZoomVirtualCanvas), new UIPropertyMetadata(null));

		public static DependencyProperty VirtualChildProperty = DependencyProperty.Register("VirtualChild", typeof(IVirtualChild), typeof(PanAndZoomVirtualCanvas));

		public event EventHandler<VisualChangeEventArgs> VisualsChanged;

		delegate void UpdateHandler();

		public PanAndZoomVirtualCanvas()
		{
			logger.Info("PanAndZoomVirtualCanvas constructor");

			_index = new QuadTree<IVirtualChild>();
			_children = new ObservableCollection<IVirtualChild>();
			_children.CollectionChanged += new System.Collections.Specialized.NotifyCollectionChangedEventHandler(OnChildrenCollectionChanged);
			Content = new Canvas();
			this.SizeChanged += new SizeChangedEventHandler(VirtualCanvas_SizeChanged);
			Content.Background = new SolidColorBrush(Colors.White);
			_backdrop = new Border();
			_backdrop.SetValue(Canvas.ZIndexProperty, int.MinValue);
			_backdrop.Background = new SolidColorBrush(Colors.White);
			Content.Children.Add(_backdrop);

			this.Children.Add(Content);
			this.Content.SetValue(Canvas.WidthProperty, double.NaN);

			this.DefaultZoomFactor = 1.4;

			Setup(this);

			// Set the Chart Geometry Clip region
			_clip = new PathGeometry();
			_clip.AddGeometry(new RectangleGeometry(new Rect(0, 0, Content.ActualWidth, Content.ActualHeight)));

			this.Clip = _clip;
		}

		void Setup(FrameworkElement control)
		{
			this.TranslateTransform = new TranslateTransform();
			this.ScaleTransform = new ScaleTransform();
			this.transformGroup = new TransformGroup();
			this.TranslateTransform.Changed += new EventHandler(OnTranslateChanged);
			this.ScaleTransform.Changed += new EventHandler(OnScaleChanged);
			this.transformGroup.Children.Add(this.ScaleTransform);
			this.transformGroup.Children.Add(this.TranslateTransform);
			this.Content.RenderTransform = this.transformGroup;
			this.Focusable = true;
			this.PreviewKeyDown += new KeyEventHandler(OnPreviewKeyDown);
			this.KeyDown += new KeyEventHandler(OnKeyDown);
			this.MouseMove += new MouseEventHandler(OnMouseMove);
			this.MouseDown += new MouseButtonEventHandler(OnMouseDown);
			this.MouseUp += new MouseButtonEventHandler(OnMouseUp);
			this.MouseWheel += new MouseWheelEventHandler(OnMouseWheel);
		}

		void OnPreviewKeyDown(object sender, KeyEventArgs e)
		{
			try
			{
				// hit escape to reset everything
				if (e.Key == Key.Escape) Reset();
			}
			catch (Exception ex)
			{
				Debug.WriteLine(ex);
				logger.Error("", ex);
			}
		}

		void OnKeyDown(object sender, KeyEventArgs e)
		{
			try
			{
				switch (e.Key)
				{
					// hit escape to reset everything
					case Key.Escape:
						Reset();
						break;

					case Key.Left:
						this.TranslateTransform.BeginAnimation(TranslateTransform.XProperty, CreatePanAnimation(1), HandoffBehavior.Compose);
						break;
				}
			}
			catch (Exception ex)
			{
				Debug.WriteLine(ex);
				logger.Error("", ex);
			}
		}

		void OnMouseWheel(object sender, MouseWheelEventArgs e)
		{
			try
			{
				// zoom into the content.  Calculate the zoom factor based on the direction of the mouse wheel.
				double zoomFactor = this.DefaultZoomFactor;
				if (e.Delta <= 0) zoomFactor = 1.0 / this.DefaultZoomFactor;
				// DoZoom requires both the logical and physical location of the mouse pointer
				var physicalPoint = e.GetPosition(this);
				DoZoom(zoomFactor, physicalPoint);
			}
			catch (Exception ex)
			{
				Debug.WriteLine(ex);
				logger.Error("", ex);
			}
		}

		void OnMouseUp(object sender, MouseButtonEventArgs e)
		{
			try
			{
				if (this.IsMouseCaptured)
				{
					// we're done.  reset the cursor and release the mouse pointer
					this.Cursor = Cursors.Arrow;
					this.ReleaseMouseCapture();
				}
			}
			catch (Exception ex)
			{
				Debug.WriteLine(ex);
				logger.Error("", ex);
			}
		}

		void OnMouseDown(object sender, MouseButtonEventArgs e)
		{
			try
			{
				// Save starting point, used later when determining how much to scroll.
				this.ScreenStartPoint = e.GetPosition(this);
				this.startOffset = new Point(this.TranslateTransform.X, this.TranslateTransform.Y);
				this.CaptureMouse();
				this.Cursor = Cursors.ScrollAll;
			}
			catch (Exception ex)
			{
				Debug.WriteLine(ex);
				logger.Error("", ex);
			}
		}

		void OnMouseMove(object sender, MouseEventArgs e)
		{
			try
			{
				if (this.IsMouseCaptured)
				{
					// if the mouse is captured then move the content by changing the translate transform.  
					// use the Pan Animation to animate to the new location based on the delta between the 
					// starting point of the mouse and the current point.
					var physicalPoint = e.GetPosition(this);
					this.TranslateTransform.BeginAnimation(TranslateTransform.XProperty, CreatePanAnimation(physicalPoint.X - this.ScreenStartPoint.X + this.startOffset.X), HandoffBehavior.Compose);
					this.TranslateTransform.BeginAnimation(TranslateTransform.YProperty, CreatePanAnimation(physicalPoint.Y - this.ScreenStartPoint.Y + this.startOffset.Y), HandoffBehavior.Compose);
					//this.TranslateTransform.X = physicalPoint.X - this.ScreenStartPoint.X + this.startOffset.X;
					//this.TranslateTransform.Y = physicalPoint.Y - this.ScreenStartPoint.Y + this.startOffset.Y;
				}
			}
			catch (Exception ex)
			{
				Debug.WriteLine(ex);
				logger.Error("", ex);
			}
		}

		/// <summary>Helper to create the panning animation for x,y coordinates.</summary>
		/// <param name="toValue">New value of the coordinate.</param>
		/// <returns>Double animation</returns>
		private DoubleAnimation CreatePanAnimation(double toValue)
		{
			var da = new DoubleAnimation(toValue, new Duration(TimeSpan.FromMilliseconds(300)));
			//da.AccelerationRatio = 0.1;
			//da.DecelerationRatio = 0.9;
			da.FillBehavior = FillBehavior.HoldEnd;
			da.Freeze();
			return da;
		}

		/// <summary>Helper to create the zoom double animation for scaling.</summary>
		/// <param name="toValue">Value to animate to.</param>
		/// <returns>Double animation.</returns>
		private DoubleAnimation CreateZoomAnimation(double toValue)
		{
			var da = new DoubleAnimation(toValue, new Duration(TimeSpan.FromMilliseconds(500)));
			//da.AccelerationRatio = 0.1;
			//da.DecelerationRatio = 0.9;
			da.FillBehavior = FillBehavior.HoldEnd;
			da.Freeze();
			return da;
		}

		/// <summary>Zoom into or out of the content.</summary>
		/// <param name="deltaZoom">Factor to mutliply the zoom level by. </param>
		/// <param name="mousePosition">Logical mouse position relative to the original content.</param>
		/// <param name="physicalPosition">Actual mouse position on the screen (relative to the parent window)</param>
		public void DoZoom(double deltaZoom, Point physicalPosition)
		{
			Point mousePosition = this.transformGroup.Inverse.Transform(physicalPosition);

			double currentZoom = this.ScaleTransform.ScaleX;
			currentZoom *= deltaZoom;
			this.TranslateTransform.BeginAnimation(TranslateTransform.XProperty, CreateZoomAnimation(-1 * (mousePosition.X * currentZoom - physicalPosition.X)));
			this.TranslateTransform.BeginAnimation(TranslateTransform.YProperty, CreateZoomAnimation(-1 * (mousePosition.Y * currentZoom - physicalPosition.Y)));
			this.ScaleTransform.BeginAnimation(ScaleTransform.ScaleXProperty, CreateZoomAnimation(currentZoom));
			this.ScaleTransform.BeginAnimation(ScaleTransform.ScaleYProperty, CreateZoomAnimation(currentZoom));
		}

		public void SetZoom(double zoom)
		{
			if (zoom <= 0)
				throw new ArgumentOutOfRangeException("zoom", zoom, "zoom must be greater than 0");

			Point physicalPosition = new Point(this.ActualWidth / 2, this.ActualHeight / 2);

			Point mousePosition = this.transformGroup.Inverse.Transform(physicalPosition);

			this.TranslateTransform.BeginAnimation(TranslateTransform.XProperty, null);
			this.TranslateTransform.BeginAnimation(TranslateTransform.YProperty, null);

			this.TranslateTransform.X = -1 * (mousePosition.X * zoom - physicalPosition.X);
			this.TranslateTransform.Y = -1 * (mousePosition.Y * zoom - physicalPosition.Y);

			this.ScaleTransform.BeginAnimation(ScaleTransform.ScaleXProperty, null);
			this.ScaleTransform.BeginAnimation(ScaleTransform.ScaleYProperty, null);

			this.ScaleTransform.ScaleX = zoom;
			this.ScaleTransform.ScaleY = zoom;

			//this.TranslateTransform.BeginAnimation(TranslateTransform.XProperty, CreateZoomAnimation(-1 * (mousePosition.X * zoom - physicalPosition.X)));
			//this.TranslateTransform.BeginAnimation(TranslateTransform.YProperty, CreateZoomAnimation(-1 * (mousePosition.Y * zoom - physicalPosition.Y)));
			//this.ScaleTransform.BeginAnimation(ScaleTransform.ScaleXProperty, CreateZoomAnimation(zoom));
			//this.ScaleTransform.BeginAnimation(ScaleTransform.ScaleYProperty, CreateZoomAnimation(zoom));
		}

		public void ZoomIn()
		{
			this.DoZoom(this.DefaultZoomFactor, new Point(this.ActualWidth / 2, this.ActualHeight / 2));
		}

		public void ZoomOut()
		{
			this.DoZoom(1.0 / this.DefaultZoomFactor, new Point(this.ActualWidth / 2, this.ActualHeight / 2));
		}

		public void PanUp()
		{
			this.TranslateTransform.BeginAnimation(TranslateTransform.YProperty, CreatePanAnimation(this.TranslateTransform.Y + 32), HandoffBehavior.Compose);
		}

		public void PanDown()
		{
			this.TranslateTransform.BeginAnimation(TranslateTransform.YProperty, CreatePanAnimation(this.TranslateTransform.Y - 32), HandoffBehavior.Compose);
		}

		public void PanLeft()
		{
			this.TranslateTransform.BeginAnimation(TranslateTransform.XProperty, CreatePanAnimation(this.TranslateTransform.X + 32), HandoffBehavior.Compose);
		}

		public void PanRight()
		{
			this.TranslateTransform.BeginAnimation(TranslateTransform.XProperty, CreatePanAnimation(this.TranslateTransform.X - 32), HandoffBehavior.Compose);
		}

		/// <summary>Reset to default zoom level and centered content.</summary>
		public void Reset()
		{
			this.TranslateTransform.BeginAnimation(TranslateTransform.XProperty, CreateZoomAnimation(0));
			this.TranslateTransform.BeginAnimation(TranslateTransform.YProperty, CreateZoomAnimation(0));
			this.ScaleTransform.BeginAnimation(ScaleTransform.ScaleXProperty, CreateZoomAnimation(1));
			this.ScaleTransform.BeginAnimation(ScaleTransform.ScaleYProperty, CreateZoomAnimation(1));
		}

		void VirtualCanvas_SizeChanged(object sender, SizeChangedEventArgs e)
		{
			try
			{
				if (_clip == null)
					_clip = new PathGeometry();

				_clip.Clear();
				_clip.AddGeometry(new RectangleGeometry(new Rect(0, 0, this.ActualWidth, this.ActualHeight)));

				this.Clip = _clip;
			}
			catch (Exception ex)
			{
				Debug.WriteLine(ex);
				logger.Error("", ex);
			}
		}

		/// <summary>
		/// Callback when _children collection is changed.
		/// </summary>
		/// <param name="sender">This</param>
		/// <param name="e">noop</param>
		void OnChildrenCollectionChanged(object sender, NotifyCollectionChangedEventArgs e)
		{
			try
			{
				RebuildVisuals();
			}
			catch (Exception ex)
			{
				Debug.WriteLine(ex);
				logger.Error("", ex);
			}
		}

		/// <summary>
		/// Returns true if all Visuals have been created for the current scroll position
		/// and there is no more idle processing needed.
		/// </summary>
		public bool IsDone
		{
			get { return _done; }
		}

		/// <summary>
		/// Resets the state so there is no Visuals associated with this canvas.
		/// </summary>
		public void RebuildVisuals()
		{
			// need to rebuild the index.
			_index = null;
			_visualPositions = null;
			_visible = Rect.Empty;
			_done = false;
			foreach (UIElement e in Content.Children)
			{
				IVirtualChild n = e.GetValue(VirtualChildProperty) as IVirtualChild;
				if (n != null)
				{
					e.ClearValue(VirtualChildProperty);
					n.DisposeVisual();
				}
			}
			Content.Children.Clear();
			Content.Children.Add(_backdrop);
			InvalidateArrange();
			StartLazyUpdate();
		}

		/// <summary>
		/// Get/Set the IVirtualChild collection.  The VirtualCanvas will call CreateVisual on them
		/// when the Bounds of your child intersects the current visible view port.
		/// </summary>
		public ObservableCollection<IVirtualChild> VirtualChildren
		{
			get { return _children; }
			set
			{
				if (value == null)
				{
					throw new ArgumentNullException("value");
				}
				if (_children != null)
				{
					_children.CollectionChanged -= new System.Collections.Specialized.NotifyCollectionChangedEventHandler(OnChildrenCollectionChanged);
				}
				_children = value;
				_children.CollectionChanged += new System.Collections.Specialized.NotifyCollectionChangedEventHandler(OnChildrenCollectionChanged);
				RebuildVisuals();
			}
		}

		/// <summary>
		/// Set the scroll amount for the scroll bar arrows.
		/// </summary>
		public Size SmallScrollIncrement
		{
			get { return _smallScrollIncrement; }
			set { _smallScrollIncrement = value; }
		}

		/// <summary>
		/// Add a new IVirtualChild.  The VirtualCanvas will call CreateVisual on them
		/// when the Bounds of your child intersects the current visible view port.
		/// </summary>
		/// <param name="c"></param>
		public void AddVirtualChild(IVirtualChild child)
		{
			_children.Add(child);
		}

		/// <summary>
		/// Return the list of virtual children that intersect the given bounds.
		/// </summary>
		/// <param name="bounds">The bounds to test</param>
		/// <returns>The list of virtual children found or null if there are none</returns>
		public IEnumerable<IVirtualChild> GetChildrenIntersecting(Rect bounds)
		{
			if (_index != null)
			{
				return _index.GetNodesInside(bounds);
			}
			return null;
		}

		/// <summary>
		/// Return true if there are any virtual children inside the given bounds.
		/// </summary>
		/// <param name="bounds">The bounds to test</param>
		/// <returns>True if a node is found whose bounds intersect the given bounds</returns>
		public bool HasChildrenIntersecting(Rect bounds)
		{
			if (_index != null)
			{
				return _index.HasNodesInside(bounds);
			}
			return false;
		}

		/// <summary>
		/// The number of visual children that are visible right now.
		/// </summary>
		public int LiveVisualCount
		{
			get { return Content.Children.Count - 1; }
		}

		/// <summary>
		/// Callback whenever the current TranslateTransform is changed.
		/// </summary>
		/// <param name="sender">TranslateTransform</param>
		/// <param name="e">noop</param>
		void OnTranslateChanged(object sender, EventArgs e)
		{
			try
			{
				OnScrollChanged();
			}
			catch (Exception ex)
			{
				Debug.WriteLine(ex);
				logger.Error("", ex);
			}
		}

		/// <summary>
		/// Callback whenever the current ScaleTransform is changed.
		/// </summary>
		/// <param name="sender">ScaleTransform</param>
		/// <param name="e">noop</param>
		void OnScaleChanged(object sender, EventArgs e)
		{
			try
			{
				OnScrollChanged();
			}
			catch (Exception ex)
			{
				Debug.WriteLine(ex);
				logger.Error("", ex);
			}
		}

		/// <summary>
		/// The ContentCanvas that is actually the parent of all the VirtualChildren Visuals.
		/// </summary>
		public Canvas ContentCanvas
		{
			get { return Content; }
		}

		/// <summary>
		/// The backgrop is the back most child of the ContentCanvas used for drawing any sort
		/// of background that is guarenteed to fill the ViewPort.
		/// </summary>
		public Border Backdrop
		{
			get { return _backdrop; }
		}

		/// <summary>
		/// Calculate the size needed to display all the virtual children.
		/// </summary>
		void CalculateExtent()
		{
			bool rebuild = false;
			if (_index == null || _extent.Width == 0 || _extent.Height == 0 ||
				 double.IsNaN(_extent.Width) || double.IsNaN(_extent.Height))
			{
				rebuild = true;
				bool first = true;
				Rect extent = new Rect();
				_visualPositions = new Dictionary<IVirtualChild, int>();
				int index = 0;
				foreach (IVirtualChild c in _children)
				{
					_visualPositions[c] = index++;

					Rect childBounds = c.Bounds;
					if (childBounds.Width != 0 && childBounds.Height != 0)
					{
						if (double.IsNaN(childBounds.Width) || double.IsNaN(childBounds.Height))
						{
							throw new InvalidOperationException(string.Format(System.Globalization.CultureInfo.CurrentUICulture,
								 "Child type '{0}' returned NaN bounds", c.GetType().Name));
						}
						if (first)
						{
							extent = childBounds;
							first = false;
						}
						else
						{
							extent = Rect.Union(extent, childBounds);
						}
					}
				}
				_extent = extent.Size;
				// Ok, now we know the size we can create the index.
				_index = new QuadTree<IVirtualChild>();
				_index.Bounds = new Rect(0, 0, extent.Width, extent.Height);
				foreach (IVirtualChild n in _children)
				{
					if (n.Bounds.Width > 0 && n.Bounds.Height > 0)
					{
						_index.Insert(n, n.Bounds);
					}
				}
			}

			// Make sure we honor the min width & height.
			double w = Math.Max(Content.MinWidth, _extent.Width);
			double h = Math.Max(Content.MinHeight, _extent.Height);
			Content.Width = w;
			Content.Height = h;

			// Make sure the backdrop covers the ViewPort bounds.
			double zoom = this.ScaleTransform.ScaleX;
			if (!double.IsInfinity(this.ViewportHeight) &&
				 !double.IsInfinity(this.ViewportHeight))
			{
				w = Math.Max(w, this.ViewportWidth / zoom);
				h = Math.Max(h, this.ViewportHeight / zoom);
				_backdrop.Width = w;
				_backdrop.Height = h;
			}

			if (_owner != null)
			{
				_owner.InvalidateScrollInfo();
			}

			if (rebuild)
			{
				AddVisibleRegion();
			}
		}

		/// <summary>
		/// WPF Measure override for measuring the control
		/// </summary>
		/// <param name="availableSize">Available size will be the viewport size in the scroll viewer</param>
		/// <returns>availableSize</returns>
		protected override Size MeasureOverride(Size availableSize)
		{
			base.MeasureOverride(availableSize);

			// We will be given the visible size in the scroll viewer here.
			CalculateExtent();

			if (availableSize != _viewPortSize)
			{
				SetViewportSize(availableSize);
			}

			foreach (UIElement child in this.InternalChildren)
			{
				IVirtualChild n = child.GetValue(VirtualChildProperty) as IVirtualChild;
				if (n != null)
				{
					Rect bounds = n.Bounds;
					child.Measure(bounds.Size);
				}
			}
			if (double.IsInfinity(availableSize.Width))
			{
				return _extent;
			}
			else
			{
				return availableSize;
			}
		}

		/// <summary>
		/// WPF ArrangeOverride for laying out the control
		/// </summary>
		/// <param name="finalSize">The size allocated by parents</param>
		/// <returns>finalSize</returns>
		protected override Size ArrangeOverride(Size finalSize)
		{
			base.ArrangeOverride(finalSize);

			CalculateExtent();

			if (finalSize != _viewPortSize)
			{
				SetViewportSize(finalSize);
			}

			Content.Arrange(new Rect(0, 0, Content.Width, Content.Height));

			if (_index == null)
			{
				StartLazyUpdate();
			}

			return finalSize;
		}

		DispatcherTimer _timer;

		/// <summary>
		/// Begin a timer for lazily creating IVirtualChildren visuals
		/// </summary>
		void StartLazyUpdate()
		{
			if (_timer == null)
			{
				_timer = new DispatcherTimer(TimeSpan.FromMilliseconds(10), DispatcherPriority.Normal,
					 new EventHandler(OnStartLazyUpdate), this.Dispatcher);
			}
			_timer.Start();
		}

		/// <summary>
		/// Callback from the DispatchTimer
		/// </summary>
		/// <param name="sender">DispatchTimer </param>
		/// <param name="args">noop</param>
		void OnStartLazyUpdate(object sender, EventArgs args)
		{
			try
			{
				_timer.Stop();
				this.LazyUpdateVisuals();
			}
			catch (Exception ex)
			{
				Debug.WriteLine(ex);
				logger.Error("", ex);
			}
		}

		/// <summary>
		/// Set the viewport size and raize a scroll changed event.
		/// </summary>
		/// <param name="s">The new size</param>
		void SetViewportSize(Size s)
		{
			if (s != _viewPortSize)
			{
				_viewPortSize = s;
				OnScrollChanged();
			}
		}

		int _createQuanta = 1000;
		int _removeQuanta = 2000;
		int _gcQuanta = 5000;
		int _idealDuration = 50; // 50 milliseconds.
		int _added;
		int _removed;
		Rect _visible = Rect.Empty;
		delegate int QuantizedWorkHandler(int quantum);

		/// <summary>
		/// Do a quantized unit of work for creating newly visible visuals, and cleaning up visuals that are no
		/// longer needed.
		/// </summary>
		void LazyUpdateVisuals()
		{
			if (_index == null)
			{
				this.CalculateExtent();
			}

			_done = true;
			_added = 0;
			_removed = 0;

			_createQuanta = SelfThrottlingWorker(_createQuanta, _idealDuration, new QuantizedWorkHandler(LazyCreateNodes));
			_removeQuanta = SelfThrottlingWorker(_removeQuanta, _idealDuration, new QuantizedWorkHandler(LazyRemoveNodes));
			_gcQuanta = SelfThrottlingWorker(_gcQuanta, _idealDuration, new QuantizedWorkHandler(LazyGarbageCollectNodes));

			if (VisualsChanged != null)
			{
				VisualsChanged(this, new VisualChangeEventArgs(_added, _removed));
			}
			if (_added > 0)
			{
				InvalidateArrange();
			}
			if (!_done)
			{
				StartLazyUpdate();
				//this.Dispatcher.BeginInvoke(DispatcherPriority.Background, new UpdateHandler(LazyUpdateVisuals));
			}
			this.InvalidateVisual();
		}

		/// <summary>
		/// Helper method for self-tuning how much time is allocated to the given handler.
		/// </summary>
		/// <param name="quantum">The current quantum allocation</param>
		/// <param name="idealDuration">The time in milliseconds we want to take</param>
		/// <param name="handler">The handler to call that does the work being throttled</param>
		/// <returns>Returns the new quantum to use next time that will more likely hit the ideal time</returns>
		private static int SelfThrottlingWorker(int quantum, int idealDuration, QuantizedWorkHandler handler)
		{
			PerfTimer timer = new PerfTimer();
			timer.Start();
			int count = handler(quantum);

			timer.Stop();
			long duration = timer.GetDuration();

			if (duration > 0 && count > 0)
			{
				long estimatedFullDuration = duration * (quantum / count);
				long newQuanta = (quantum * idealDuration) / estimatedFullDuration;
				quantum = Math.Max(100, (int)Math.Min(newQuanta, int.MaxValue));
			}

			return quantum;
		}

		/// <summary>
		/// Create visuals for the nodes that are now visible.
		/// </summary>
		/// <param name="quantum">Amount of work we can do here</param>
		/// <returns>Amount of work we did</returns>
		private int LazyCreateNodes(int quantum)
		{

			if (_visible == Rect.Empty)
			{
				_visible = GetVisibleRect();
				_visibleRegions.Add(_visible);
				_done = false;
			}

			int count = 0;
			int regionCount = 0;
			while (_visibleRegions.Count > 0 && count < quantum)
			{
				Rect r = _visibleRegions[0];
				_visibleRegions.RemoveAt(0);
				regionCount++;

				// Iterate over the visible range of nodes and make sure they have visuals.
				foreach (IVirtualChild n in _index.GetNodesInside(r))
				{
					if (n.Visual == null)
					{
						EnsureVisual(n);
						_added++;
					}

					count++;

					if (count >= quantum)
					{
						// This region is too big, so subdivide it into smaller slices.
						if (regionCount == 1)
						{
							// We didn't even complete 1 region, so we better split it.
							SplitRegion(r, _visibleRegions);
						}
						else
						{
							_visibleRegions.Add(r); // put it back since we're not done!
						}
						_done = false;
						break;
					}
				}

			}
			return count;
		}

		/// <summary>
		/// Insert the visual for the child in the same order as is is defined in the 
		/// VirtualChildren collection so the visuals draw on top of each other in the expected order.
		/// The trick is that GetNodesIntersecting returns the nodes in pretty much random order depending 
		/// on how the QuadTree decides to break up the canvas.  
		/// 
		/// The thing we should avoid is a linear search through the potentially large collection of 
		/// IVirtualChildren to compute its visible index which is why we have the _visualPositions map.  
		/// We should also avoid a N*M algorithm where N is the number of nodes returned from GetNodesIntersecting 
		/// and M is the number of children already visible.  For example, Page down in a zoomed out situation 
		/// gives potentially high N and and M which would basically be an O(n2) algorithm.  
		/// 
		/// So the solution is to use the _visualPositions map to get the expected visual position index
		/// of a given IVirtualChild, then do a binary search through existing visible children to find the
		/// insertion point of the new child.  So this is O(Log M).  
		/// </summary>
		/// <param name="child">The IVirtualChild to add visual for</param>
		public void EnsureVisual(IVirtualChild child)
		{
			if (child.Visual != null)
			{
				return;
			}

			UIElement e = child.CreateVisual();
			e.SetValue(VirtualChildProperty, child);
			Rect bounds = child.Bounds;
			Canvas.SetLeft(e, bounds.Left);
			Canvas.SetTop(e, bounds.Top);

			if (_visualPositions == null)
				return;

			// Get the correct absolute position of this child.
			int position = _visualPositions[child];

			// Now do a binary search for the correct insertion position based
			// on the visual positions of the existing visible children.
			UIElementCollection c = Content.Children;
			int min = 0;
			int max = c.Count - 1;
			while (max > min + 1)
			{
				int i = (min + max) / 2;
				UIElement v = Content.Children[i];
				IVirtualChild n = v.GetValue(VirtualChildProperty) as IVirtualChild;
				if (n != null)
				{
					int index = _visualPositions[n];
					if (index > position)
					{
						// search from min to i.
						max = i;
					}
					else
					{
						// search from i to max.
						min = i;
					}
				}
				else
				{
					// Any nodes without IVirtualChild should be behind the
					// IVirtualChildren by definition (like the Backdrop).
					min = i;
				}
			}

			// If 'max' is the last child in the collection, then we need to see
			// if we have a new last child.
			if (max == c.Count - 1)
			{
				UIElement v = c[max];
				IVirtualChild maxchild = v.GetValue(VirtualChildProperty) as IVirtualChild;
				int maxpos = position;
				if (maxchild == null || position > _visualPositions[maxchild])
				{
					// Then we have a new last child!
					max++;
				}
			}

			c.Insert(max, e);

		}


		/// <summary>
		/// Split a rectangle into 2 and add them to the regions list.
		/// </summary>
		/// <param name="r">Rectangle to split</param>
		/// <param name="regions">List to add to</param>
		private void SplitRegion(Rect r, IList<Rect> regions)
		{
			double minWidth = this.SmallScrollIncrement.Width * 2;
			double minHeight = this.SmallScrollIncrement.Height * 2;

			if (r.Width > r.Height && r.Height > minHeight)
			{
				// horizontal slices
				double h = r.Height / 2;
				regions.Add(new Rect(r.Left, r.Top, r.Width, h + 10));
				regions.Add(new Rect(r.Left, r.Top + h, r.Width, h + 10));
			}
			else if (r.Width < r.Height && r.Width > minWidth)
			{
				// vertical slices
				double w = r.Width / 2;
				regions.Add(new Rect(r.Left, r.Top, w + 10, r.Height));
				regions.Add(new Rect(r.Left + w, r.Top, w + 10, r.Height));
			}
			else
			{
				regions.Add(r); // put it back since we're not done!
			}
		}

		/// <summary>
		/// Remove visuals for nodes that are no longer visible.
		/// </summary>
		/// <param name="quantum">Amount of work we can do here</param>
		/// <returns>Amount of work we did</returns>
		private int LazyRemoveNodes(int quantum)
		{
			Rect visible = GetVisibleRect();
			int count = 0;

			// Also remove nodes that are no longer visible.
			int regionCount = 0;
			while (_dirtyRegions.Count > 0 && count < quantum)
			{
				int last = _dirtyRegions.Count - 1;
				Rect dirty = _dirtyRegions[last];
				_dirtyRegions.RemoveAt(last);
				regionCount++;

				// Iterate over the visible range of nodes and make sure they have visuals.
				foreach (IVirtualChild n in _index.GetNodesInside(dirty))
				{
					UIElement e = n.Visual;
					if (e != null)
					{
						Rect nrect = n.Bounds;
						if (!nrect.IntersectsWith(visible))
						{
							e.ClearValue(VirtualChildProperty);
							Content.Children.Remove(e);
							n.DisposeVisual();
							_removed++;
						}
					}

					count++;
					if (count >= quantum)
					{
						if (regionCount == 1)
						{
							// We didn't even complete 1 region, so we better split it.
							SplitRegion(dirty, _dirtyRegions);
						}
						else
						{
							_dirtyRegions.Add(dirty); // put it back since we're not done!
						}
						_done = false;
						break;
					}
				}
			}
			return count;
		}

		/// <summary>
		/// Check all child nodes to see if any leaked from LazyRemoveNodes and remove their visuals.
		/// </summary>
		/// <param name="quantum">Amount of work we can do here</param>
		/// <returns>The amount of work we did</returns>
		int LazyGarbageCollectNodes(int quantum)
		{

			int count = 0;
			// Now after every update also do a full incremental scan over all the children
			// to make sure we didn't leak any nodes that need to be removed.
			while (count < quantum && _nodeCollectCycle < Content.Children.Count)
			{
				UIElement e = Content.Children[_nodeCollectCycle++];
				IVirtualChild n = e.GetValue(VirtualChildProperty) as IVirtualChild;
				if (n != null)
				{
					Rect nrect = n.Bounds;
					if (!nrect.IntersectsWith(_visible))
					{
						e.ClearValue(VirtualChildProperty);
						Content.Children.Remove(e);
						n.DisposeVisual();
						_removed++;
					}
					count++;
				}
				_nodeCollectCycle++;
			}

			if (_nodeCollectCycle < Content.Children.Count)
			{
				_done = false;
			}

			return count;
		}

#if DEBUG_DUMP
		public void ShowQuadTree(bool show)
		{
			if (show)
			{
				_index.ShowQuadTree(Content);
			}
			else
			{
				RebuildVisuals();
			}
		}

		public void Dump(string fileName)
		{
			using (StreamWriter w = new StreamWriter(fileName))
			{
				using (LogWriter log = new LogWriter(w))
				{
					log.Open("QuadTree");
					_index.Dump(log);
					log.Open("Other");
					log.WriteAttribute("MaxDepth", log.MaxDepth.ToString(CultureInfo.CurrentUICulture));
					log.Close();
					log.Close();
				}
			}
		}
#endif

		/// <summary>
		/// Return the full size of this canvas.
		/// </summary>
		public Size Extent
		{
			get { return _extent; }
		}

		#region IScrollInfo Members

		/// <summary>
		/// Return whether we are allowed to scroll horizontally.
		/// </summary>
		public bool CanHorizontallyScroll
		{
			get { return _canHScroll; }
			set { _canHScroll = value; }
		}

		/// <summary>
		/// Return whether we are allowed to scroll vertically.
		/// </summary>
		public bool CanVerticallyScroll
		{
			get { return _canVScroll; }
			set { _canVScroll = value; }
		}

		/// <summary>
		/// The height of the canvas to be scrolled.
		/// </summary>
		public double ExtentHeight
		{
			get { return _extent.Height * this.ScaleTransform.ScaleY; }
		}

		/// <summary>
		/// The width of the canvas to be scrolled.
		/// </summary>
		public double ExtentWidth
		{
			get { return _extent.Width * this.ScaleTransform.ScaleX; }
		}

		/// <summary>
		/// Scroll down one small scroll increment.
		/// </summary>
		public void LineDown()
		{
			SetVerticalOffset(VerticalOffset + (_smallScrollIncrement.Height * this.ScaleTransform.ScaleX));
		}

		/// <summary>
		/// Scroll left by one small scroll increment.
		/// </summary>
		public void LineLeft()
		{
			SetHorizontalOffset(HorizontalOffset - (_smallScrollIncrement.Width * this.ScaleTransform.ScaleX));
		}

		/// <summary>
		/// Scroll right by one small scroll increment
		/// </summary>
		public void LineRight()
		{
			SetHorizontalOffset(HorizontalOffset + (_smallScrollIncrement.Width * this.ScaleTransform.ScaleX));
		}

		/// <summary>
		/// Scroll up by one small scroll increment
		/// </summary>
		public void LineUp()
		{
			SetVerticalOffset(VerticalOffset - (_smallScrollIncrement.Height * this.ScaleTransform.ScaleX));
		}

		/// <summary>
		/// Make the given visual at the given bounds visible.
		/// </summary>
		/// <param name="visual">The visual that will become visible</param>
		/// <param name="rectangle">The bounds of that visual</param>
		/// <returns>The bounds that is actually visible.</returns>
		public Rect MakeVisible(System.Windows.Media.Visual visual, Rect rectangle)
		{
			return rectangle;
		}

		/// <summary>
		/// Scroll down by one mouse wheel increment.
		/// </summary>
		public void MouseWheelDown()
		{
			SetVerticalOffset(VerticalOffset + (_smallScrollIncrement.Height * this.ScaleTransform.ScaleX));
		}

		/// <summary>
		/// Scroll left by one mouse wheel increment.
		/// </summary>
		public void MouseWheelLeft()
		{
			SetHorizontalOffset(HorizontalOffset + (_smallScrollIncrement.Width * this.ScaleTransform.ScaleX));
		}

		/// <summary>
		/// Scroll right by one mouse wheel increment.
		/// </summary>
		public void MouseWheelRight()
		{
			SetHorizontalOffset(HorizontalOffset - (_smallScrollIncrement.Width * this.ScaleTransform.ScaleX));
		}

		/// <summary>
		/// Scroll up by one mouse wheel increment.
		/// </summary>
		public void MouseWheelUp()
		{
			SetVerticalOffset(VerticalOffset - (_smallScrollIncrement.Height * this.ScaleTransform.ScaleX));
		}

		/// <summary>
		/// Page down by one view port height amount.
		/// </summary>
		public void PageDown()
		{
			SetVerticalOffset(VerticalOffset + _viewPortSize.Height);
		}

		/// <summary>
		/// Page left by one view port width amount.
		/// </summary>
		public void PageLeft()
		{
			SetHorizontalOffset(HorizontalOffset - _viewPortSize.Width);
		}

		/// <summary>
		/// Page right by one view port width amount.
		/// </summary>
		public void PageRight()
		{
			SetHorizontalOffset(HorizontalOffset + _viewPortSize.Width);
		}

		/// <summary>
		/// Page up by one view port height amount.
		/// </summary>
		public void PageUp()
		{
			SetVerticalOffset(VerticalOffset - _viewPortSize.Height);
		}

		/// <summary>
		/// Return the ScrollViewer that contains this object.
		/// </summary>
		public ScrollViewer ScrollOwner
		{
			get { return _owner; }
			set { _owner = value; }
		}

		/// <summary>
		/// Scroll to the given absolute horizontal scroll position.
		/// </summary>
		/// <param name="offset">The horizontal position to scroll to</param>
		public void SetHorizontalOffset(double offset)
		{
			double xoffset = Math.Max(Math.Min(offset, ExtentWidth - ViewportWidth), 0);
			this.TranslateTransform.X = -xoffset;
			OnScrollChanged();
		}

		/// <summary>
		/// Scroll to the given absolute vertical scroll position.
		/// </summary>
		/// <param name="offset">The vertical position to scroll to</param>
		public void SetVerticalOffset(double offset)
		{
			double yoffset = Math.Max(Math.Min(offset, ExtentHeight - ViewportHeight), 0);
			this.TranslateTransform.Y = -yoffset;
			OnScrollChanged();
		}

		/// <summary>
		/// Get the current horizontal scroll position.
		/// </summary>
		public double HorizontalOffset
		{
			get { return -this.TranslateTransform.X; }
		}

		/// <summary>
		/// Return the current vertical scroll position.
		/// </summary>
		public double VerticalOffset
		{
			get { return -this.TranslateTransform.Y; }
		}

		/// <summary>
		/// Return the height of the current viewport that is visible in the ScrollViewer.
		/// </summary>
		public double ViewportHeight
		{
			get { return _viewPortSize.Height; }
		}

		/// <summary>
		/// Return the width of the current viewport that is visible in the ScrollViewer.
		/// </summary>
		public double ViewportWidth
		{
			get { return _viewPortSize.Width; }
		}

		#endregion

		/// <summary>
		/// Get the currently visible rectangle according to current scroll position and zoom factor and
		/// size of scroller viewport.
		/// </summary>
		/// <returns>A rectangle</returns>
		Rect GetVisibleRect()
		{
			// Add a bit of extra around the edges so we are sure to create nodes that have a tiny bit showing.
			double xstart = (this.HorizontalOffset - _smallScrollIncrement.Width) / this.ScaleTransform.ScaleX;
			double ystart = (this.VerticalOffset - _smallScrollIncrement.Height) / this.ScaleTransform.ScaleY;
			double xend = (this.HorizontalOffset + (_viewPortSize.Width + (2 * _smallScrollIncrement.Width))) / this.ScaleTransform.ScaleX;
			double yend = (this.VerticalOffset + (_viewPortSize.Height + (2 * _smallScrollIncrement.Height))) / this.ScaleTransform.ScaleY;
			return new Rect(xstart, ystart, xend - xstart, yend - ystart);
		}

		/// <summary>
		/// The visible region has changed, so we need to queue up work for dirty regions and new visible regions
		/// then start asynchronously building new visuals via StartLazyUpdate.
		/// </summary>
		void OnScrollChanged()
		{
			Rect dirty = _visible;
			AddVisibleRegion();
			_nodeCollectCycle = 0;
			_done = false;

			Rect intersection = Rect.Intersect(dirty, _visible);
			if (intersection == Rect.Empty)
			{
				_dirtyRegions.Add(dirty); // the whole thing is dirty
			}
			else
			{
				// Add left stripe
				if (dirty.Left < intersection.Left)
				{
					_dirtyRegions.Add(new Rect(dirty.Left, dirty.Top, intersection.Left - dirty.Left, dirty.Height));
				}
				// Add right stripe
				if (dirty.Right > intersection.Right)
				{
					_dirtyRegions.Add(new Rect(intersection.Right, dirty.Top, dirty.Right - intersection.Right, dirty.Height));
				}
				// Add top stripe
				if (dirty.Top < intersection.Top)
				{
					_dirtyRegions.Add(new Rect(dirty.Left, dirty.Top, dirty.Width, intersection.Top - dirty.Top));
				}
				// Add right stripe
				if (dirty.Bottom > intersection.Bottom)
				{
					_dirtyRegions.Add(new Rect(dirty.Left, intersection.Bottom, dirty.Width, dirty.Bottom - intersection.Bottom));
				}
			}

			StartLazyUpdate();
			InvalidateScrollInfo();
		}

		/// <summary>
		/// Tell the ScrollViewer to update the scrollbars because, extent, zoom or translate has changed.
		/// </summary>
		public void InvalidateScrollInfo()
		{
			if (_owner != null)
			{
				_owner.InvalidateScrollInfo();
			}
		}

		/// <summary>
		/// Add the current visible rect to the list of regions to process
		/// </summary>
		private void AddVisibleRegion()
		{
			_visible = GetVisibleRect();
			_visibleRegions.Clear();
			_visibleRegions.Add(_visible);
		}


#if DEBUG_QUAD_TREE
        public void ShowQuadTree(bool show)
        {
            if (show)
            {
                _index.ShowQuadTree(_content);
            }
            else
            {
                RebuildVisuals();
            }
        }

        public void Dump(string filename)
        {
            using (StreamWriter w = new StreamWriter(filename))
            {
                using (LogWriter log = new LogWriter(w))
                {
                    log.Open("QuadTree");
                    _index.Dump(log);
                    log.Open("Other");
                    log.WriteAttribute("MaxDepth", log.MaxDepth.ToString());
                    log.Close();
                    log.Close();
                }
            }
        }
#endif

	}
}
